use crate::algorithm::alpha_beta::AlphaBeta;
use crate::algorithm::pvs::PrincipalVariationSearch;
use crate::algorithm::Algorithm;
use crate::game::fox_and_hounds::{FHMove, FoxAndHounds, Position};
use crate::game::Game;

fn set_board_state(
    model: &mut FoxAndHounds,
    move_no: i32,
    is_fox_next: bool,
    fox_pos: i32,
    hounds_pos: [i32; 4],
) {
    model.move_no = move_no;
    model.is_fox_next = is_fox_next;
    model.fox_pos = Position(fox_pos as usize);
    model.hounds_pos = [
        Position(hounds_pos[0] as usize),
        Position(hounds_pos[1] as usize),
        Position(hounds_pos[2] as usize),
        Position(hounds_pos[3] as usize),
    ];
    model.reset_free_squares();
}

fn mov(mo: i32) -> Option<FHMove> {
    Some(FHMove::new_h(
        (mo / 100) as usize,
        Position((mo % 100) as usize),
    ))
}

fn do_move(model: &mut FoxAndHounds, mo: i32) {
    *model = model.execute_move(mov(mo).unwrap());
}

fn do_ai_move<A: Algorithm>(model: &mut FoxAndHounds, depth: i32, do_precalculated_moves: bool) {
    *model = model.execute_move(
        model
            .calculate_move::<A>(depth, do_precalculated_moves)
            .unwrap(),
    );
}

#[test]
fn it_works() {
    assert!(true);
}

#[test]
#[should_panic(expected = "expected panic string")]
fn should_panic() {
    panic!("expected panic string");
}

#[test]
fn model_initialization() {
    let m = FoxAndHounds::new();
    assert_eq!(m.is_fox_next, true);
    assert_eq!(m.fox_pos, Position(15));
    assert_eq!(m.hounds_pos[0], Position(82));
    assert_eq!(m.hounds_pos[1], Position(84));
    assert_eq!(m.hounds_pos[2], Position(86));
    assert_eq!(m.hounds_pos[3], Position(88));
    assert_eq!(m.is_game_over(), false);
}

#[test]
fn moves() {
    let mut m = FoxAndHounds::new();
    do_move(&mut m, 424);
    assert_eq!(m.is_fox_next, false);
    assert_eq!(m.fox_pos, Position(24));
    assert_eq!(m.hounds_pos[0], Position(82));
    assert_eq!(m.hounds_pos[1], Position(84));
    assert_eq!(m.hounds_pos[2], Position(86));
    assert_eq!(m.hounds_pos[3], Position(88));
    assert_eq!(m.is_game_over(), false);
}

#[test]
#[should_panic(expected = "fox cannot reach that square")]
fn illegal_move_1() {
    let mut m = FoxAndHounds::new();
    do_move(&mut m, 415);
}

#[test]
#[should_panic(expected = "Index not initialized correctly for fox move")]
fn illegal_move_2() {
    let mut m = FoxAndHounds::new();
    do_move(&mut m, 524);
}

#[test]
#[should_panic(expected = "invalid index")]
fn illegal_move_3() {
    let mut m = FoxAndHounds::new();
    do_move(&mut m, 424);
    do_move(&mut m, 524);
}

#[test]
fn fox_wins() {
    let mut m = FoxAndHounds::new();
    do_move(&mut m, 424);
    do_move(&mut m, 071);
    do_move(&mut m, 433);
    do_move(&mut m, 275);
    do_move(&mut m, 442);
    do_move(&mut m, 377);
    do_move(&mut m, 453);
    do_move(&mut m, 062);
    do_move(&mut m, 464);
    do_move(&mut m, 051);
    do_move(&mut m, 473);
    do_move(&mut m, 042);
    do_move(&mut m, 482);
    /*
     *  F H _ _
     * _ _ H H
     *  _ _ _ _
     * _ _ _ _
     *  H _ _ _
     * _ _ _ _
     *  _ _ _ _
     * _ _ _ _
     */
    assert_eq!(m.is_fox_next, false);
    assert_eq!(m.fox_pos, Position(82));
    assert_eq!(m.hounds_pos[0], Position(42));
    assert_eq!(m.hounds_pos[1], Position(84));
    assert_eq!(m.hounds_pos[2], Position(75));
    assert_eq!(m.hounds_pos[3], Position(77));
    assert_eq!(m.is_game_over(), true);
}

#[test]
fn hounds_win() {
    let mut m = FoxAndHounds::new();
    do_move(&mut m, 424);
    do_move(&mut m, 071);
    do_move(&mut m, 413);
    do_move(&mut m, 062);
    do_move(&mut m, 422);
    do_move(&mut m, 051);
    do_move(&mut m, 411);
    do_move(&mut m, 042);
    do_move(&mut m, 422);
    do_move(&mut m, 031);
    do_move(&mut m, 411);
    do_move(&mut m, 022);
    /*
     *  _ H H H
     * _ _ _ _
     *  _ _ _ _
     * _ _ _ _
     *  _ _ _ _
     * _ _ _ _
     *  H _ _ _
     * F _ _ _
     */
    assert_eq!(m.is_fox_next, true);
    assert_eq!(m.fox_pos, Position(11));
    assert_eq!(m.hounds_pos[0], Position(22));
    assert_eq!(m.hounds_pos[1], Position(84));
    assert_eq!(m.hounds_pos[2], Position(86));
    assert_eq!(m.hounds_pos[3], Position(88));
    assert_eq!(m.is_game_over(), true);
}

#[test]
fn hounds_win_after_calculated_move() {
    let mut m = FoxAndHounds::new();
    set_board_state(&mut m, 11, false, 11, [33, 84, 86, 88]);
    do_ai_move::<PrincipalVariationSearch>(&mut m, 10, true);
    /*
     *  _ H H H
     * _ _ _ _
     *  _ _ _ _
     * _ _ _ _
     *  _ _ _ _
     * _ _ _ _
     *  H _ _ _
     * F _ _ _
     */
    assert_eq!(m.is_fox_next, true);
    assert_eq!(m.fox_pos, Position(11));
    assert_eq!(m.hounds_pos[0], Position(22));
    assert_eq!(m.hounds_pos[1], Position(84));
    assert_eq!(m.hounds_pos[2], Position(86));
    assert_eq!(m.hounds_pos[3], Position(88));
    assert_eq!(m.is_game_over(), true);
}

#[test]
fn fox_wins_after_calculated_move() {
    let mut m = FoxAndHounds::new();
    set_board_state(&mut m, 12, true, 71, [22, 84, 86, 88]);
    do_ai_move::<PrincipalVariationSearch>(&mut m, 10, true);
    /*
     * .F.H.H.H
     * _._._._.
     * ._._._._
     * _._._._.
     * ._._._._
     * _._._._.
     * .H._._._
     * _._._._.
     */
    assert_eq!(m.is_fox_next, false);
    assert_eq!(m.fox_pos, Position(82));
    assert_eq!(m.hounds_pos[0], Position(22));
    assert_eq!(m.hounds_pos[1], Position(84));
    assert_eq!(m.hounds_pos[2], Position(86));
    assert_eq!(m.hounds_pos[3], Position(88));
    assert_eq!(m.is_game_over(), true);
}

#[test]
fn fox_shall_move_forwards() {
    let mut m = FoxAndHounds::new();
    set_board_state(&mut m, 3, true, 22, [71, 84, 86, 88]);
    /*
     * ._.H.H.H
     * H._._._.
     * ._._._._
     * _._._._.
     * ._._._._
     * _._._._.
     * .F._._._
     * _._._._.
     */
    assert_eq!(m.is_fox_next, true);
    assert_eq!(m.fox_pos, Position(22));
    assert_eq!(m.hounds_pos[0], Position(71));
    assert_eq!(m.hounds_pos[1], Position(84));
    assert_eq!(m.hounds_pos[2], Position(86));
    assert_eq!(m.hounds_pos[3], Position(88));
    assert_eq!(m.is_game_over(), false);

    do_ai_move::<PrincipalVariationSearch>(&mut m, 10, true);
    /*
     * ._.H.H.H
     * H._._._.
     * ._._._._
     * _._._._.
     * ._._._._
     * _.F._._.
     * ._._._._
     * _._._._.
     */
    assert_eq!(m.is_fox_next, false);
    assert_eq!(m.fox_pos, Position(33));
    assert_eq!(m.hounds_pos[0], Position(71));
    assert_eq!(m.hounds_pos[1], Position(84));
    assert_eq!(m.hounds_pos[2], Position(86));
    assert_eq!(m.hounds_pos[3], Position(88));
    assert_eq!(m.is_game_over(), false);
}

#[test]
fn end_game() {
    let mut m = FoxAndHounds::new();
    set_board_state(&mut m, 48, false, 15, [13, 33, 35, 26]);
    do_ai_move::<PrincipalVariationSearch>(&mut m, 10, true);
    /*
     * ._._._._      ._._._._
     * _._._._.      _._._._.
     * ._._._._      ._._._._
     * _._._._.  ->  _._._._.
     * ._._._._      ._._._._
     * _.H.H._.      _._.H._.
     * ._._.H._      ._.H.H._
     * _.H.F._.      _.H.F._.
     */
    print!("{:?}, F: {:?}, H: {:?}", m, m.fox_pos, m.hounds_pos);
    assert_eq!(m.is_game_over(), true);
    // either hound can move to position 24
}

#[test]
fn middle_game() {
    let mut m = FoxAndHounds::new();
    set_board_state(&mut m, 14, false, 15, [51, 73, 75, 77]);
    do_ai_move::<PrincipalVariationSearch>(&mut m, 12, true);
    /*
     * ._._._._      ._._._._
     * _.H.H.H.      _.H.H._.
     * ._._._._      ._._._.H
     * H._._._.  ->  H._._._.
     * ._._._._      ._._._._
     * _._._._.      _._._._.
     * ._._._._      ._._._._
     * _._.F._.      _._.F._.
     */
    print!("{:?}", m);
    assert_eq!(m.is_game_over(), false);
}

#[test]
fn precomputed_moves() {
    let mut m = FoxAndHounds::new();
    for _ in 0..10 {
        do_ai_move::<PrincipalVariationSearch>(&mut m, 8, true);
    }
    m = FoxAndHounds::new();
    m.move_no = 8;
    for _ in 0..10 {
        do_ai_move::<PrincipalVariationSearch>(&mut m, 8, true);
    }
}

#[test]
fn test_ai_games() {
    let mut m = FoxAndHounds::new_fox_starts(true);
    while !m.is_game_over() {
        do_ai_move::<PrincipalVariationSearch>(&mut m, 8, true);
    }
    m = FoxAndHounds::new_fox_starts(false);
    while !m.is_game_over() {
        do_ai_move::<PrincipalVariationSearch>(&mut m, 8, true);
    }
}

#[test]
fn test_ai_games_2() {
    let mut m = FoxAndHounds::new_fox_starts(true);
    while !m.is_game_over() {
        do_ai_move::<AlphaBeta>(&mut m, 8, true);
    }
    m = FoxAndHounds::new_fox_starts(false);
    while !m.is_game_over() {
        do_ai_move::<AlphaBeta>(&mut m, 8, true);
    }
}

#[test]
#[should_panic(expected = "new position already occupied")]
fn illegal_move_4() {
    let m = FoxAndHounds::new();
    m.check_fox_move(&FHMove::new_f(Position(82)));
}

#[test]
#[should_panic(expected = "fox moves outside board")]
fn illegal_move_5() {
    let m = FoxAndHounds::new();
    m.check_fox_move(&FHMove::new_f(Position(06)));
}
