use crate::Algorithm;
use crate::Game;
use crate::Move;
use std::fmt;

#[derive(Debug, Copy, Clone, PartialEq, Eq)]
pub struct NCMove(usize);

impl Move for NCMove {
    fn to_str(&self) -> String {
        match self.0 {
            0 => "Top Left",
            1 => "Top Middle",
            2 => "Top Right",
            3 => "Middle Left",
            4 => "Center",
            5 => "Middle Right",
            6 => "Bottom Left",
            7 => "Bottom Middle",
            8 => "Bottom Right",
            _ => panic!("Internal error: invalid position {}", self.0)
        }.to_string()
    }
}

#[derive(Clone)]
pub struct NoughtsAndCrosses {
    player_one: bool,
    fields: [usize; 9],
    move_no: i32,
}

impl fmt::Debug for NoughtsAndCrosses {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        writeln!(f, "+---+---+---+")?;
        for y in 0..3 {
            write!(f, "| ")?;
            for x in 0..3 {
                let pos = y * 3 + x;
                if self.fields[pos] == 1 {
                    write!(f, "O")?;
                } else if self.fields[pos] == 2 {
                    write!(f, "X")?;
                } else {
                    write!(f, " ")?;
                }
                write!(f, " | ")?;
            }
            writeln!(f, "\n+---+---+---+")?;
        }
        Ok(())
    }
}

impl Game for NoughtsAndCrosses {
    type Move = NCMove;

    fn possible_moves(&self) -> Vec<Self::Move> {
        if self.player_won(1) || self.player_won(2) {
            return vec![];
        }
        let mut res = Vec::<NCMove>::with_capacity(9);
        for i in 0..9 {
            if self.fields[i] == 0 {
                res.push(NCMove(i));
            }
        }
        res
    }
    fn execute_move(&self, mov: Self::Move) -> Self {
        let mut next = self.clone();
        if self.possible_moves().contains(&mov) {
            next.do_move(mov);
        } else {
            panic!("Invalid move {:?}", mov);
        }
        next
    }
    fn do_move(&mut self, mov: Self::Move) -> Self::Move {
        self.fields[mov.0] = if self.player_one { 1 } else { 2 };
        self.move_no += 1;
        self.player_one = !self.player_one;
        mov
    }
    fn undo_move(&mut self, mov: Self::Move) {
        self.fields[mov.0] = 0;
        self.move_no -= 1;
        self.player_one = !self.player_one;
    }
    fn get_score(&self) -> i32 {
        if self.is_game_over() {
            if self.player_won(1) && self.player_one || self.player_won(2) && !self.player_one {
                self.move_no - 100
            } else {
                100 - self.move_no
            }
        } else {
            0
        }
    }
    fn score_if_win_next_round(&self) -> i32 {
        -self.move_no - 1
    }
    fn is_game_over(&self) -> bool {
        self.possible_moves().is_empty()
    }
    fn calculate_move<T: Algorithm>(&self, depth: i32, _: bool) -> Option<Self::Move> {
        T::compute(self, depth)
    }
}

impl NoughtsAndCrosses {
    pub fn new() -> NoughtsAndCrosses {
        NoughtsAndCrosses {
            player_one: true,
            fields: [0; 9],
            move_no: 0,
        }
    }
    fn player_won(&self, p: usize) -> bool {
        self.fields[0] == p && self.fields[1] == p && self.fields[2] == p
            || self.fields[3] == p && self.fields[4] == p && self.fields[5] == p
            || self.fields[6] == p && self.fields[7] == p && self.fields[8] == p
            || self.fields[0] == p && self.fields[3] == p && self.fields[6] == p
            || self.fields[1] == p && self.fields[4] == p && self.fields[7] == p
            || self.fields[2] == p && self.fields[5] == p && self.fields[8] == p
            || self.fields[0] == p && self.fields[4] == p && self.fields[8] == p
            || self.fields[2] == p && self.fields[4] == p && self.fields[6] == p
    }
}
